@@ -3,17 +3,10 @@
 /// \file       private.h
 /// \brief      Common includes, definions, and prototypes
 //
-//  Copyright (C) 2007 Lasse Collin
+//  Author:     Lasse Collin
 //
-//  This program is free software; you can redistribute it and/or
-//  modify it under the terms of the GNU Lesser General Public
-//  License as published by the Free Software Foundation; either
-//  version 2.1 of the License, or (at your option) any later version.
-//
-//  This program is distributed in the hope that it will be useful,
-//  but WITHOUT ANY WARRANTY; without even the implied warranty of
-//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-//  Lesser General Public License for more details.
+//  This file has been put into the public domain.
+//  You can do whatever you want with this file.
 //
 ///////////////////////////////////////////////////////////////////////////////
 
@@ -51,11 +44,11 @@
 #endif
 
 #include "main.h"
-#include "process.h"
+#include "coder.h"
 #include "message.h"
 #include "args.h"
 #include "hardware.h"
-#include "io.h"
+#include "file_io.h"
 #include "options.h"
 #include "signals.h"
 #include "suffix.h"
